use std::borrow::Cow;
use std::fmt::Debug;

use serde::{Deserialize, Serialize};
use url::Url;

use crate::link::Link;

pub mod link;
pub mod object;
pub mod recipe;

pub trait ObjectTyped: Clone + Debug + Sized + 'static {
    fn get_type() -> &'static str;
}

#[macro_export]
macro_rules! def_ld {
    ($obj_type: expr, $x: ty) => {
        impl $crate::ObjectTyped for $x {
            fn get_type() -> &'static str {
                $obj_type
            }
        }
    };
}

#[macro_export]
macro_rules! ld_union {
    ($z: ident, $($item_name: ident as $item_type: ty),+) => {

        #[derive(Clone, Debug, Serialize, Deserialize)]
        pub struct $z {
            $(
            #[serde(flatten)]
            $item_name: Box<$item_type>,
            )+
        }
    };
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum OneOrMore<T> {
    One(T),
    Many(Vec<T>),
}

impl<T> OneOrMore<T> {
    pub fn into_vec(self) -> Vec<T> {
        match self {
            OneOrMore::One(x) => vec![x],
            OneOrMore::Many(x) => x,
        }
    }

    pub fn iter(&self) -> std::slice::Iter<'_, T> {
        match self {
            OneOrMore::One(x) => std::slice::from_ref(x).iter(),
            OneOrMore::Many(x) => x.iter(),
        }
    }
}

#[derive(Clone, Debug, Default, Serialize, Deserialize)]
pub struct Base {
    #[serde(rename = "@type")]
    as_type: Option<OneOrMore<String>>,
    #[serde(rename = "@id")]
    as_id: Option<String>,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum LinkOrUrl<'a> {
    Url(Url),
    Link(Box<Link<'a>>),
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[repr(transparent)]
pub struct Id(pub String);

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum ObjectSlot<'a, T: ObjectTyped> {
    Id(Id),
    Value(Cow<'a, T>),
}
